from datetime import datetime
from re import fullmatch, search
from urllib.parse import ParseResult, quote_plus
from urllib.request import urlopen
from uuid import uuid5, NAMESPACE_URL, UUID

from bs4 import BeautifulSoup, Tag
from weasyprint import CSS, HTML

from . import Scraper
from ..generator import PackageData

class DeviantArt(Scraper):
	domain = 'deviantart.com'

	@staticmethod
	def scrape(url: ParseResult) -> PackageData:
		_STORE_NAME = 'DeviantArt'
		_STORE_PREFIX = 'DVNT'

		request = urlopen(f"https://backend.deviantart.com/oembed?format=xml&url={quote_plus(url.geturl())}")
		soup = BeautifulSoup(request.read(), 'xml')

		if (type := soup.select_one('type')) and type.get_text() == 'link' and (link_url := soup.select_one('url')):
			canonical_url = link_url.get_text()
		else:
			canonical_url = url.geturl()

		sku = int(sku_match.group(1)) if (sku_match := search(r'-(\d+)$', canonical_url)) else None
		name = title.get_text() if (title := soup.select_one('title')) else None
		artists = [author.get_text()] if (author := soup.select_one('author_name')) else []
		pubdate = pubdate.get_text() if (pubdate := soup.select_one('pubdate')) else None
		product_image = image.get_text() if (image := soup.select_one('fullsize_url')) else None

		if product_image:
			soup = BeautifulSoup(urlopen(canonical_url).read(), 'lxml')
		else:
			product_image = image.get_text() if (image := soup.select_one('url')) else None
			soup = BeautifulSoup(urlopen(url.geturl()).read(), 'lxml')

		da_app_url = da_app_element.attrs['content'] if isinstance(da_app_element := soup.find('meta', attrs={'property': 'da:appurl'}), Tag) else None
		global_id = UUID(uuid_match.group(1)) if da_app_url and (uuid_match := fullmatch(r'DeviantArt://deviation/([0-9A-Z]{8}-[0-9A-Z]{4}-[0-9A-Z]{4}-[0-9A-Z]{4}-[0-9A-Z]{12})', da_app_url)) else uuid5(NAMESPACE_URL, canonical_url)
		description = desc_element if (desc_element := soup.select_one('#description div')) else None

		html = BeautifulSoup('', 'lxml')
		stylesheet = CSS(string='@page { margin: 1em; } img { max-width: 100%; }')

		h1 = soup.new_tag(name='h1')	# pyright: ignore[reportUnknownMemberType]
		h1.string = name or 'Unknown Product Name'
		html.append(h1)

		if artists:
			h2 = soup.new_tag(name='h2')	# pyright: ignore[reportUnknownMemberType]
			h2.string = f"By {artists[0]}"
			html.append(h2)

		if pubdate:
			h3 = soup.new_tag(name='h3')	# pyright: ignore[reportUnknownMemberType]
			h3.string = f"Published {datetime.fromisoformat(pubdate).strftime('%Y-%m-%d')}"
			html.append(h3)

		if description:
			html.append(soup.new_tag('hr'))	# pyright: ignore[reportUnknownMemberType]
			html.append(description)
			description = description.get_text('\n')
		else:
			description = None

		if product_image:
			html.append(soup.new_tag('hr'))	# pyright: ignore[reportUnknownMemberType]
			img = soup.new_tag('img')  # pyright: ignore[reportUnknownMemberType]
			img.attrs['src'] = product_image
			html.append(img)
			product_image = urlopen(product_image)
		else:
			product_image = None

		return PackageData(
			global_id = global_id,
			prefix = _STORE_PREFIX,
			store = _STORE_NAME,
			sku = sku,
			name = name,
			artists = artists,
			description = description,
			image = product_image,
			readme = HTML(string=html.decode_contents()).write_pdf(stylesheets=[stylesheet])
		)

		return PackageData()


from abc import ABC, abstractmethod
from re import search, IGNORECASE
from typing import Any, Self
from urllib.parse import ParseResult, urlparse

from ..generator import PackageData

class Scraper(ABC):
	domain: str
	scrapers: dict[str, type[Self]] = {}

	def __init_subclass__(cls, **kwargs: dict[str, Any]) -> None:
		if not hasattr(cls, 'domain'):
			raise TypeError(f'{cls.__name__}.domain must be set.')
		super().__init_subclass__(**kwargs)
		cls.scrapers[cls.domain] = cls

	@staticmethod
	@abstractmethod
	def scrape(url: ParseResult) -> PackageData:
		raise NotImplementedError

def get_scraper(url: ParseResult | None) -> type[Scraper] | None:
	if url and url.hostname:
		if url.hostname == 'web.archive.org' and (suburl := search('(https?://.*)', url.path, IGNORECASE)):
			url = urlparse(suburl.group(0))

		if url.hostname and (site := url.hostname.split("www.")[-1]) in Scraper.scrapers:
			return Scraper.scrapers[site]
		else:
			raise ValueError(f'No scraper for site: {url.hostname}')

def scrape(url: ParseResult) -> PackageData:
	try:
		if scraper := get_scraper(url):
			return scraper.scrape(url)
	except ValueError:
		pass

	return PackageData()


import json

from re import search
from typing import cast
from urllib.parse import ParseResult, urlparse
from urllib.request import urlopen
from uuid import uuid5, NAMESPACE_URL

from bs4 import BeautifulSoup, Tag
from weasyprint import CSS, HTML

from . import Scraper
from ..generator import PackageData

class RenderHub(Scraper):
	domain = 'renderhub.com'

	@staticmethod
	def scrape(url: ParseResult) -> PackageData:
		_STORE_NAME = 'RenderHub'
		_STORE_PREFIX = 'RNDRHUB'

		request = urlopen(url.geturl())
		actual_url = urlparse(str(request.url))
		soup = BeautifulSoup(request.read(), 'lxml')

		base_url = base_tag.attrs['href'] if (base_tag := soup.select_one('base[href]')) else f'{ actual_url.scheme }://{ actual_url.netloc }'
		canonical_url = url_element.attrs['content'] if isinstance(url_element := soup.find('meta', attrs={'property': 'og:url'}), Tag) else None

		if canonical_url:
			global_id = uuid5(NAMESPACE_URL, canonical_url)

			ratings_script = ratings_script.get_text() if (ratings_script := soup.select_one('div.itemInfo script:-soup-contains("function showRat(p)")')) else None
			sku = int(sku_match.group(1)) if ratings_script and (sku_match := search(r'"i=(\d+)&p="', ratings_script)) else None

			pubDate_tag = pubLabel.parent.next_sibling if (pubLabel := soup.select_one('div.itemInfo table.infoTable b:-soup-contains("Published:")')) else None	# pyright: ignore[reportOptionalMemberAccess]
			pubdate = pubDate_tag.get_text() if pubDate_tag else None

			name = img.attrs['title'] if (img := soup.select_one('#mainImg')) else None
			description = soup.select_one('div.prodInfo')

			ldjson: list[dict[str, str]] | None = json.loads(ldjson_str.text) if (ldjson_str := soup.select_one('script[type="application/ld+json"]')) else None
			if ldjson and (data := next(filter(lambda x: isinstance(x, dict) and x.get('@type') == 'Product', ldjson), None)):	# pyright: ignore[reportUnnecessaryIsInstance]
				if not name:
					name = data.get('name', None)

				artists = [artist] if (artist := cast(dict[str, dict[str, str]], data).get('brand', {}).get('name', None)) else []
				images = cast(dict[str, list[str]], data).get('image', [])

				html = BeautifulSoup('', 'lxml')
				stylesheet = CSS(string='@page { margin: 1em; } img { max-width: 100%; }')

				h1 = soup.new_tag(name='h1')	# pyright: ignore[reportUnknownMemberType]
				h1.string = name or 'Unknown Product Name'
				html.append(h1)

				if artists:
					h2 = soup.new_tag(name='h2')	# pyright: ignore[reportUnknownMemberType]
					h2.string = f"By {artists[0]}"
					html.append(h2)

				if pubdate:
					h3 = soup.new_tag(name='h3')	# pyright: ignore[reportUnknownMemberType]
					h3.string = f"Published {pubdate}"
					html.append(h3)

				if description:
					html.append(soup.new_tag('hr'))	# pyright: ignore[reportUnknownMemberType]
					html.append(description)
					description = description.get_text('\n')
				else:
					description = None

				if images:
					product_image = urlopen(images[0])
					html.append(soup.new_tag('hr'))	# pyright: ignore[reportUnknownMemberType]
					for image in images:
						img = soup.new_tag('img')  # pyright: ignore[reportUnknownMemberType]
						img.attrs['src'] = image
						html.append(img)
				else:
					product_image = None

				return PackageData(
					global_id = global_id,
					prefix = _STORE_PREFIX,
					store = _STORE_NAME,
					sku = sku,
					name = name,
					artists = artists,
					description = description,
					image = product_image,
					readme = HTML(string=html.decode_contents(), base_url=base_url).write_pdf(stylesheets=[stylesheet])
				)

		return PackageData()


import warnings

from pathlib import Path

from lxml import etree

from .packagedata import AssetData, ObjectData, PackageData

def load_metadata(content_location: Path) -> PackageData:
	_SUPPORT_DIR = 'Runtime/Support'

	metadata_files: list[Path] = []
	package_data = PackageData()

	if (support_path := content_location / _SUPPORT_DIR).is_dir():
		for entry in support_path.iterdir():
			if entry.is_file() and entry.suffix == '.dsx':
				metadata_files.append(entry)

	if len(metadata_files) > 1:
		raise RuntimeError('Multiple metadata files not supported by this tool')
    
	if metadata_files:
		try:
			metadata = etree.parse(metadata_files[0])
		except Exception as e:
			warnings.warn(f'Invalid Metadata file: {e}', RuntimeWarning)
		else:
			for e in metadata.xpath('/ContentDBInstall/Products/Product'):
				package_data.name = e.attrib['VALUE']
				package_data.global_id = eGUID[0].attrib['VALUE'] if (eGUID := e.xpath('GlobalID')) else None
				package_data.store = eStore[0].attrib['VALUE'] if (eStore := e.xpath('StoreID')) else None
				package_data.sku = int(eToken[0].attrib['VALUE']) if (eToken := e.xpath('ProductToken')) else None
				package_data.description = eDescription[0].text if (eDescription := e.xpath('Description')) else None
				package_data.artists = [a.attrib['VALUE'] for a in e.xpath('Artists/Artist')]
				package_data.objects = set(ObjectData(o.attrib['VALUE'], o.attrib['REF']) for o in e.xpath('ObjectCompatibilities/ObjectCompatibility'))
				package_data.assets = {Path(a.attrib['VALUE']).as_posix(): AssetData(
					aDescription[0].text if (aDescription := a.xpath('Description')) else None,
					aContentType[0].attrib['VALUE'] if (aContentType := a.xpath('ContentType')) else None,
					aAudience[0].attrib['VALUE'] if (aAudience := a.xpath('Audience')) else None,
					[c.attrib['VALUE'] for c in a.xpath('Categories/Category[string-length(@VALUE) > 0]')],
					[c.attrib['VALUE'] for c in a.xpath('Compatibilities/Compatibility[string-length(@VALUE) > 0]')],
					aCompatibilityBase[0].attrib['VALUE'] if (aCompatibilityBase := a.xpath('CompatibilityBase')) else None,
					[t.attrib['VALUE'] for t in a.xpath('Tags/Tag[string-length(@VALUE) > 0]')],
					[u.attrib['VALUE'] for u in a.xpath('Userwords/Userword[string-length(@VALUE) > 0]')],
					aUserNotes[0].text if (aUserNotes := a.xpath('UserNotes')) else None
				) for a in e.xpath('Assets/Asset')}

		if ((img := metadata_files[0].with_suffix('.jpg')).exists() or
			(img := metadata_files[0].with_suffix('.png')).exists()):
			package_data.image = img

		if package_data.store == 'LOCAL USER':
			package_data.store = None

	return package_data


import gzip
import re
import sys
import zipfile

from dataclasses import dataclass, fields, KW_ONLY
from http.client import HTTPResponse
from mimetypes import guess_extension
from pathlib import Path
from struct import pack
from urllib.parse import ParseResult
from uuid import uuid4

from lxml import etree

from .loadmetadata import load_metadata
from .packagedata import PackageData
from ..scrapers import scrape

_CONTENT_DIR = 'Content'
_SUPPORT_DIR = 'Runtime/Support'

_USER_FACING_DAZ_EXTENSIONS = ('.daz', '.djl', '.duf', '.ds', '.dsa', '.dsb', '.dse')
_OTHER_DAZ_EXTENSIONS = ('.dhdm', '.dsd', '.dsf', '.dsj', '.dso', '.dsv')
_COMPRESSABLE_DAZ_EXTENSIONS = ('.dsf', '.duf')
_NON_USER_FACING_DIRECTORIES = ('data', "readme's", 'runtime', 'uninstallers')

_USER_FACING_POSER_EXTENSIONS = ('.pz3', '.pzz', '.cr2', '.crz', '.pz2', '.p2z', '.fc2', '.fcz', '.hr2', '.hrz', '.hd2', '.hdz', '.pp2', '.ppz', '.lt2', '.ltz', '.cm2', '.cmz', '.mc6', '.mcz', '.mt5', '.mz5')
_OTHER_POSER_EXTENSIONS = ('.pmd',)
_POSER_USER_FACING_DIRECTORIES = ('camera', 'character', 'collections', 'face', 'hair', 'hand', 'light', 'materials', 'pose', 'props', 'scene')

_EXCLUDE_FILES = (x.lower() for x in ['.DS_Store', '._.DS_Store', 'InstallManagerFileRegister.json', 'Desktop.ini', 'pspbrwse.jbf', 'Thumbs.db'])
_EXCLUDE_SUFFIXES = ('.xmp',)


# DIM does not support Unicode filenames in ZIP files due to using the minizip 1.01 library.
# Examining official DAZ packages reveals Windows-1252 encoding is used. Packages with
# filenames using non ASCII characters may not extract correctly on OS X.
Ascii_ZipInfo = zipfile.ZipInfo
class Windows1252_ZipInfo(zipfile.ZipInfo):
	def _encodeFilenameFlags(self):
		try:
			return self.filename.encode('Windows-1252'), self.flag_bits
		except UnicodeEncodeError as e:
			raise SystemExit(f'DIM does not support Unicode filenames: {self.filename}') from e

	def __init__(self, filename: str = "NoName", date_time: tuple[int, int, int, int, int, int] = (1980,1,1,0,0,0)):
		super().__init__(filename, date_time)

		if self.filename.encode() != self.filename.encode('Windows-1252', 'replace'):
			UNICODE_PATH_EXTRA_FIELD_ID = 0x7075
			encoded_filename = self.filename.encode()
			crc = zipfile.crc32(self._encodeFilenameFlags()[0])	# pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType, reportUnknownVariableType]
			self.extra += pack(f'<HHBL{len(encoded_filename)}s', UNICODE_PATH_EXTRA_FIELD_ID, len(encoded_filename) + 5, 1, crc, encoded_filename)

@dataclass
class PackageGenerator(PackageData):
	id: int | None = None
	url: ParseResult | None = None
	verbose: bool = False
	_: KW_ONLY
	content_location: Path

	@property
	def sanitized_name(self) -> str:
		alpha_numeric_regex = re.compile(r'[^0-9A-Za-z]')
		return alpha_numeric_regex.sub('', self.name or '')

	@property
	def zip_filename(self) -> str:
		# Max SKU length for filename is 8 characters, but can be longer in metadata
		# Overflow SKU into source prefix if space is available as SKU will be incorrect
		# in DIM either way, and DIM does not allow filtering by source prefix
		sku = f'{self.sku:08}'[-8:]
		prefix = f'{self.prefix}{str(self.sku)[:-8][:7 - len(self.prefix or '')]}'
		return f"{prefix}{sku}{f'-{self.id:02}' if self.id else ''}_{self.sanitized_name}.zip"

	@property
	def metadata_filename_base(self) -> str:
		# NOTE: This is a guess and may need to be expanded with additional characters
		_REPLACE_METADATA_FILENAME_CHARACTERS = r'[\\/*?:"<>|!. ]'
		return re.sub(_REPLACE_METADATA_FILENAME_CHARACTERS, '_', f"{self.store}_{self.sku}_{self.name}")

	@property
	def metadata_image_extension(self) -> str | None:
		extension = None

		if isinstance(self.image, Path):
			extension = self.image.suffix
		elif isinstance(self.image, HTTPResponse) and self.image:
			# strict=False because ShareCG specifies incorrect MIME type for JPEG.
			if ext := guess_extension(self.image.info()['Content-Type'], False):
				extension = ext

		return extension if extension in ['.jpg', '.png'] else None

	@property
	def manifest_file(self) -> bytes:
		if self.verbose:
			print('Generating Manifest')

		manifest = etree.Element('DAZInstallManifest', VERSION='0.1')
		etree.SubElement(manifest, 'GlobalID', VALUE=str(self.global_id))

		for _, relative_file_path in self._files():
			etree.SubElement(manifest, 'File', TARGET='Content', ACTION='Install', VALUE=(_CONTENT_DIR / relative_file_path).as_posix())

		etree.SubElement(manifest, 'File', TARGET='Content', ACTION='Install', VALUE=f'{_CONTENT_DIR}/{_SUPPORT_DIR}/{self.metadata_filename_base}.dsa')
		etree.SubElement(manifest, 'File', TARGET='Content', ACTION='Install', VALUE=f'{_CONTENT_DIR}/{_SUPPORT_DIR}/{self.metadata_filename_base}.dsx')
		if image_extension := self.metadata_image_extension:
			etree.SubElement(manifest, 'File', TARGET='Content', ACTION='Install', VALUE=f'{_CONTENT_DIR}/{_SUPPORT_DIR}/{self.metadata_filename_base}{image_extension}')

		etree.indent(manifest, space=' ')
		return etree.tostring(manifest, encoding='UTF-8', pretty_print=True)

	@property
	def supplement_file(self) -> bytes:
		if self.verbose:
			print('Generating Supplement')

		if not self.name:
			raise RuntimeError(f'Missing Product Name')

		supplement = etree.Element('ProductSupplement', VERSION='0.1')
		etree.SubElement(supplement, 'ProductName', VALUE=self.name)
		etree.SubElement(supplement, 'InstallTypes', VALUE='Content')
		etree.SubElement(supplement, 'ProductTags', VALUE=','.join(sorted(self.tags)))

		etree.indent(supplement, space=' ')
		return etree.tostring(supplement, encoding='UTF-8', pretty_print=True)

	@property
	def metadata_file(self) -> bytes:
		if self.verbose:
			print('Generating Metadata')

		if not self.name:
			raise RuntimeError(f'Missing Product Name')

		metadata = etree.Element('ContentDBInstall', VERSION='1.0')
		product = etree.SubElement(etree.SubElement(metadata, 'Products'), 'Product', VALUE=self.name)

		if self.store:
			etree.SubElement(product, 'StoreID', VALUE=self.store)

		if self.global_id:
			etree.SubElement(product, 'GlobalID', VALUE=str(self.global_id))

		if self.sku:
			etree.SubElement(product, 'ProductToken', VALUE=str(self.sku))

		if self.description:
			etree.SubElement(product, 'Description').text=etree.CDATA(self.description)

		if self.artists:
			artists = etree.SubElement(product, 'Artists')
			for artist in self.artists:
				etree.SubElement(artists, 'Artist', VALUE=artist)

		# Assets / SupportAssets
		assets = etree.SubElement(product, 'Assets')
		support_assets = etree.SubElement(product, 'SupportAssets', VALUE=f'/{_SUPPORT_DIR}/{self.metadata_filename_base}.dsx')
		for _, relative_file_path, in self._files():
			suffix = relative_file_path.suffix.lower()
			path_parts = [p.lower() for p in relative_file_path.parts]
			if ((suffix in _USER_FACING_DAZ_EXTENSIONS and not path_parts[0] in _NON_USER_FACING_DIRECTORIES) or
				(suffix in _USER_FACING_DAZ_EXTENSIONS + _USER_FACING_POSER_EXTENSIONS and len(path_parts) > 3 and
	 			path_parts[0] == 'runtime' and path_parts[1] == 'libraries' and path_parts[2] in _POSER_USER_FACING_DIRECTORIES)):

				asset_path = relative_file_path.as_posix()
				asset = etree.SubElement(assets, 'Asset', VALUE=asset_path)

				if self.assets and asset_path in self.assets:
					asset_data = self.assets[asset_path]
					if asset_data.description:
						etree.SubElement(asset, 'Description').text=etree.CDATA(asset_data.description)

					if asset_data.content_type:
						etree.SubElement(asset, 'ContentType', VALUE=asset_data.content_type)

					if asset_data.audience:
						etree.SubElement(asset, 'Audience', VALUE=asset_data.audience)

					if asset_data.categories and len(asset_data.categories):
						categories = etree.SubElement(asset, 'Categories')
						for category in asset_data.categories:
							etree.SubElement(categories, 'Category', VALUE=category)

					if asset_data.tags and len(asset_data.tags):
						tags = etree.SubElement(asset, 'Tags')
						for tag in asset_data.tags:
							etree.SubElement(tags, 'Tag', VALUE=tag)

					if asset_data.compatibilities and len(asset_data.compatibilities):
						compatibilities = etree.SubElement(asset, 'Compatibilities')
						for compatibility in asset_data.compatibilities:
							etree.SubElement(compatibilities, 'Compatibility', VALUE=compatibility)

					if asset_data.compatibility_base:
						etree.SubElement(asset, 'CompatibilityBase', VALUE=asset_data.compatibility_base)

					if asset_data.user_words and len(asset_data.user_words):
						userwords = etree.SubElement(asset, 'Userwords')
						for userword in asset_data.user_words:
							etree.SubElement(userwords, 'Userword', VALUE=userword)

					if asset_data.user_notes:
						etree.SubElement(asset, 'UserNotes').text=etree.CDATA(asset_data.user_notes or '')

			else:
				etree.SubElement(support_assets, 'SupportAsset', VALUE=('/' / relative_file_path).as_posix())

		if self.objects:
			objectcompatibilities = etree.SubElement(product, 'ObjectCompatibilities')
			for object in self.objects:
				etree.SubElement(objectcompatibilities, 'ObjectCompatibility', VALUE=object.uri, REF=object.scene_id)

		etree.indent(metadata, space=' ')
		return etree.tostring(metadata, encoding='UTF-8', xml_declaration=True, pretty_print=True)

	@property
	def metadata_script(self) -> str:
		return '''// DAZ Studio version 0.0.0.0 filetype DAZ Script

if( App.version >= 67109158 ) //4.0.0.294
{
	var oFile = new DzFile( getScriptFileName() );
	var oAssetMgr = App.getAssetMgr();
	if( oAssetMgr )
	{
		oAssetMgr.queueDBMetaFile( oFile.baseName() );
	}
}
'''

	@property
	def guess_tags(self) -> set[str]:
		tags: set[str] = set()

		if any(x for _, x in self._files() if x.suffix.lower() in _USER_FACING_DAZ_EXTENSIONS + _OTHER_DAZ_EXTENSIONS and not x.parent.as_posix().lower() == 'runtime/support'):
			tags.add('DAZStudio4_5')

		if any(x for _, x in self._files() if x.suffix.lower() in _USER_FACING_POSER_EXTENSIONS + _OTHER_POSER_EXTENSIONS and x.as_posix().lower().startswith('runtime/')):
			tags.add('DAZStudio4_5')
			tags.add('PoserLegacy')

		return tags or set(('General',))

	def __post_init__(self):
		self._merge_package_data(load_metadata(self.content_location))

		if self.url:
			self._merge_package_data(scrape(self.url))

		if not self.global_id:
			self.global_id = uuid4()

		if not self.prefix:
			self.prefix = 'LOCAL'

		if not self.store:
			self.store = 'LOCAL USER'

		if not self.tags:
			self.tags = self.guess_tags

	def _merge_package_data(self, package_data: PackageData) -> None:
		for field in fields(PackageData):
			if (not getattr(self, field.name)):
				setattr(self, field.name, getattr(package_data, field.name))

	def _assert_required_vars(self) -> None:
		requiredString: list[str] = []
		exceptions: list[Exception] = []

		if not self.prefix:
			requiredString.append('Source Prefix')

		if not self.sku:
			requiredString.append('Product SKU')

		if not self.name:
			requiredString.append('Product Name')

		if requiredString:
			exceptions.append(RuntimeError(f'Missing {', '.join(requiredString)}'))

		if not self.sanitized_name:
			exceptions.append(RuntimeError(f"Unable to generate sanitized product name from '{self.name}'"))

		if exceptions:
			raise ExceptionGroup('Required variables missing', exceptions)

	def _files(self):
		for root, _, files in self.content_location.walk():
			if not (rel_root := root.relative_to(self.content_location)).as_posix() == _SUPPORT_DIR:
				for file in files:
					path = Path(file)
					if not '__MACOSX' in path.parts and not path.name.lower() in _EXCLUDE_FILES and not path.suffix.lower() in _EXCLUDE_SUFFIXES:
						file_path = root / file
						if not (path.suffix.lower() == '.rsr' and file_path.with_suffix('.png').exists()):
							yield(file_path, rel_root / file)

	def _create_zip(self) -> None:
		_MANIFEST_FILE = 'Manifest.dsx'
		_SUPPLEMENT_FILE = 'Supplement.dsx'
		_README_FILE = 'ReadMe.pdf'

		if self.verbose:
			print('Generating ZIP file')

		self._assert_required_vars()

		# Use Windows-1252 encoding for filenames
		zipfile.ZipInfo = Windows1252_ZipInfo

		with zipfile.ZipFile(self.zip_filename, 'x', zipfile.ZIP_DEFLATED) as zip_file:
			# DIM Package
			zip_file.writestr(_MANIFEST_FILE, self.manifest_file)
			zip_file.writestr(_SUPPLEMENT_FILE, self.supplement_file)

			# ReadMe
			if isinstance(self.readme, Path):
				zip_file.write(self.readme, _README_FILE)
			elif isinstance(self.readme, bytes):
				zip_file.writestr(_README_FILE, self.readme)

			# Metadata
			zip_file.writestr(f'{_CONTENT_DIR}/{_SUPPORT_DIR}/{self.metadata_filename_base}.dsx', self.metadata_file)
			zip_file.writestr(f'{_CONTENT_DIR}/{_SUPPORT_DIR}/{self.metadata_filename_base}.dsa', self.metadata_script)
			if (image_extension := self.metadata_image_extension):
				if isinstance(self.image, Path):
					zip_file.write(self.image, f'{_CONTENT_DIR}/{_SUPPORT_DIR}/{self.metadata_filename_base}{image_extension}')
				elif isinstance(self.image, HTTPResponse) and self.image:
					zip_file.writestr(f'{_CONTENT_DIR}/{_SUPPORT_DIR}/{self.metadata_filename_base}{image_extension}', self.image.read())

			# Content
			if self.verbose:
				print('Adding Content')

			for file_path, relative_file_path, in self._files():
				# Compress DAZ compressable files before adding to archive
				if file_path.suffix.lower() in _COMPRESSABLE_DAZ_EXTENSIONS and file_path.stat().st_size:
					with gzip.open(file_path) as f:
						try:
							f.read(1)
							zip_file.write(file_path, _CONTENT_DIR / relative_file_path)
						except gzip.BadGzipFile:
							if self.verbose:
								print (f'Compressing: {relative_file_path}')
							with open(file_path, 'rb') as f:
								zip_file.writestr(zipfile.ZipInfo.from_file(file_path, (_CONTENT_DIR / relative_file_path).as_posix()), gzip.compress(f.read()))
				else:
					zip_file.write(file_path, _CONTENT_DIR / relative_file_path)

		# Revert filename encoding
		zipfile.ZipInfo = Ascii_ZipInfo

	def make_package(self) -> None:
		try:
			self._create_zip()
		except (FileExistsError) as e:
			sys.exit(e)	# pyright: ignore[reportArgumentType]


from dataclasses import dataclass, field
from http.client import HTTPResponse
from pathlib import Path
from uuid import UUID

@dataclass
class AssetData:
    description: str | None = None
    content_type: str | None = None
    audience: str | None = None
    categories: list[str] = field(default_factory=list)
    compatibilities: list[str] = field(default_factory=list)
    compatibility_base: str | None = None
    tags: list[str] = field(default_factory=list)
    user_words: list[str] = field(default_factory=list)
    user_notes: str | None = None
    
@dataclass(frozen=True)
class ObjectData:
    uri: str
    scene_id: str

@dataclass
class PackageData:
    global_id: UUID | None = None
    prefix: str | None = None
    sku: int | None = None
    store: str | None = None
    name: str | None = None
    tags: set[str] = field(default_factory=set)
    artists: list[str] = field(default_factory=list)
    description: str | None = None
    image: Path | HTTPResponse | None = None
    readme: bytes | Path | None = None
    assets: dict[str, AssetData] = field(default_factory=dict)
    objects: set[ObjectData] = field(default_factory=set)


from re import search
from typing import cast
from urllib.parse import ParseResult, quote_plus, urlparse
from urllib.request import urlopen
from uuid import uuid5, NAMESPACE_URL

from bs4 import BeautifulSoup, Tag
from weasyprint import CSS, HTML

from . import Scraper
from ..generator import PackageData

class Renderosity(Scraper):
	domain = 'renderosity.com'

	@staticmethod
	def scrape(url: ParseResult) -> PackageData:
		_STORE_NAME = 'Renderosity'
		_STORE_PREFIX = 'ROSITY'

		_FREE_NAME = 'Renderosity Free'
		_FREE_PREFIX = 'ROSITYF'

		request = urlopen(url.geturl())
		actual_url = urlparse(str(request.url))
		soup = BeautifulSoup(request.read(), 'lxml')

		base_url = base_tag.attrs['href'] if (base_tag := soup.select_one('base[href]')) else f'{ actual_url.scheme }://{ actual_url.netloc }'
		page_type = type_element.attrs['content'] if isinstance(type_element := soup.find('meta', attrs={'property': 'og:type'}), Tag) else None
		canonical_url = url_element.attrs['content'] if isinstance(url_element := soup.find('meta', attrs={'property': 'og:url'}), Tag) else None

		if canonical_url and page_type in ('item', 'product'):
			html = BeautifulSoup('', 'lxml')
			stylesheet = CSS(string='@page { margin: 1em; } img { max-width: 100%; } @page my_page { size: landscape; } .rr-mkt-product-image { page: my_page; }')
			global_id = uuid5(NAMESPACE_URL, canonical_url)
			description = description_element.attrs['content'] if isinstance(description_element := soup.find('meta', attrs={'property': 'og:description'}), Tag) else None
			product_image = urlopen(img_element.attrs['content']) if isinstance(img_element := soup.find('meta', attrs={'property': 'og:image'}), Tag) and img_element.attrs['content'] else None

			if url.path.lower().startswith('/freestuff/'):
				if soup.find('meta', attrs={'property': 'og:type', 'content': 'item'}):

					sku = int(sku_match.group(0)) if (sku_match := search(r'(\d+)', canonical_url)) else None
					name = title_element.attrs['content'] if isinstance(title_element := soup.find('meta', attrs={'property': 'og:title'}), Tag) else None
					artists = [a.get_text() for a in soup.select('.rr-fsitem-byline > a')]

					if product_html := soup.select_one('div.rr-fsitem'):
						if downloads := product_html.select_one('span.rr-fsitem-downloads'):
							downloads.replace_with(soup.new_tag('hr'))  # pyright: ignore[reportUnknownMemberType]
						if section := product_html.select_one('span.rr-fsitem-section'):
							for a in section.findAll('a'):
								a.replaceWithChildren()
						html.append(product_html)

					if not product_image and name:
						search_request = urlopen(f'https://www.renderosity.com/users/{quote_plus(artists[0])}/freestuff?keyword={quote_plus(name)}')
						search_soup = BeautifulSoup(search_request.read(), 'lxml')
						product_image = urlopen(img_element.attrs['src']) if (img_element := search_soup.select_one(f'div[data-id="{sku}"] img.rr-fsitem-tile-thumbnail')) and img_element.attrs['src'] else None

					return PackageData(
						global_id = global_id,
						prefix = _FREE_PREFIX,
						store = _FREE_NAME,
						sku = sku,
						name = name,
						artists = artists,
						description = description,
						image = product_image,
						readme = HTML(string=html.decode_contents(), base_url=base_url).write_pdf(stylesheets=[stylesheet])
					)
			else:
				if soup.find('meta', attrs={'property': 'og:type', 'content': 'product'}):

					sku = int(sku_element.attrs['data-id']) if (sku_element := soup.select_one('div.rr-mkt-product-container')) else None
					name = name_element.get_text() if (name_element := soup.select_one('span.rr-mktproduct-title')) else None
					artists = [a.get_text() for a in soup.select('.rr-mktproduct-byline > a')]

					if product_html := soup.select_one('div.rr-item-view'):
						if img := product_html.select_one('div.rr-mkt-product-imageline'):
							img.replace_with(soup.new_tag('hr'))  # pyright: ignore[reportUnknownMemberType]
						if video := product_html.select_one('div.rr-mkt-product-360-video-wrapper'):
							video.decompose()
						for cover in product_html.select('div.rr-nudity-cover'):
							cover.decompose()
						for sidebar in product_html.select('div.rr-mktproduct-sidebar'):
							sidebar.decompose()
						if departments := product_html.select_one('div.rr-mkt-product-departments'):
							departments.decompose()

						if info_wrapper := product_html.select_one('div.rr-mkt-product-additional_info-wrapper'):
							for element in info_wrapper.find_next_siblings():
								cast(Tag, element).decompose()
							info_wrapper.extract()
							if (info_url := info_url.attrs['data-additional_info_url'] if (info_url := info_wrapper.select_one('ul.nav')) else None):
								for pane in info_wrapper.select('a.nav-link'):
									if (href := pane['href']) in ('#description', '#editorial'):
										pane_content = BeautifulSoup(urlopen(f'{ info_url }?pane={ href.lstrip('#') }').read(), 'lxml')
#										if href == '#editorial':
#											x = pane_content.select_one('body > table table:nth-child(12)').decompose()

										if pane_content and (pane_content.find('img') or pane_content.get_text(strip=True) != ''):
											product_html.append(soup.new_tag('hr'))  # pyright: ignore[reportUnknownMemberType]
											header = soup.new_tag('h1')  # pyright: ignore[reportUnknownMemberType]
											header.append(pane.decode_contents())
											product_html.append(header)
											product_html.append(pane_content)

						html.append(product_html)

					return PackageData(
						global_id = global_id,
						prefix = _STORE_PREFIX,
						store = _STORE_NAME,
						sku = sku,
						name = name,
						artists = artists,
						description = description,
						image = product_image,
						readme = HTML(string=html.decode_contents(), base_url=base_url).write_pdf(stylesheets=[stylesheet])
					)

		return PackageData()


from re import search, IGNORECASE
from urllib.parse import ParseResult, urljoin, urlparse, parse_qs
from urllib.request import urlopen, Request
from uuid import uuid5, NAMESPACE_URL

from bs4 import BeautifulSoup, Tag
from weasyprint import CSS, HTML

from . import Scraper
from ..generator import PackageData

class Renderotica(Scraper):
	domain = 'renderotica.com'

	@staticmethod
	def scrape(url: ParseResult) -> PackageData:
		_STORE_NAME = 'Renderotica'
		_STORE_PREFIX = 'ROTICA'

		request = urlopen(Request(url.geturl(), headers={'User-Agent': 'Mozilla'}))
		actual_url = urlparse(str(request.url))
		soup = BeautifulSoup(request.read(), 'lxml')

		base_url = base_tag.attrs['href'] if (base_tag := soup.select_one('base[href]')) else f'{ actual_url.scheme }://{ actual_url.netloc }'

		if actual_url.hostname == 'web.archive.org' and (suburl := search('(https?://.*)', url.path, IGNORECASE)):
			canonical_url = urlparse(suburl.group(0)).geturl().split('_', 1)[0] + '_'
		else:
			canonical_url = actual_url.geturl().split('_', 1)[0] + '_'

		global_id = uuid5(NAMESPACE_URL, canonical_url)
		sku = int(sku_match.group(1)) if (sku_match := search(r'/sku/(\d+)', canonical_url)) else None

		
		if (product_listing := soup.select_one('#product-listing')) and (product_info := product_listing.select_one('div.product-info')):
			product_content = soup.select_one('div.product-listing-content')
			product_content_old = soup.select_one('#product-info-panes')

			if product_content or product_content_old:
				html = BeautifulSoup('', 'lxml')
				stylesheet = CSS(string='''
					@page { margin: 1em; }
					img { max-width: 100%; }
					.ProductCompatibility td.level2,
					.ProductCompatibility td.level3,
					.ProductCompatibility td.level4
						{ background-color: transparent; }
				''')
				name = name_element.get_text() if (name_element := product_info.select_one('.product-name')) else None
				description = description_element.get_text() if (description_element := product_info.select_one('.summary')) else None
				product_image = urljoin(base_url, img_element.attrs['src']) if (img_element := product_listing.select_one('.primary-image img[src]')) else None
				artists = [vendor_element.get_text() for vendor_element in  product_info.select('a.vendor')]

				if pinterest := product_info.select_one('#p_lt_zoneContent_pageplaceholder_p_lt_zoneContentBody_CGBytes_ProductDetail_ctl01_hlkPinterest[href]'):
					product_image = parse_qs(urlparse(pinterest.attrs['href']).query)['media'][0]

				if description_element and (sku_element := product_info.select_one('#p_lt_zoneContent_pageplaceholder_p_lt_zoneContentBody_CGBytes_ProductDetail_lblSkuNumber[title]')):
					released_date = soup.new_tag('div')	# pyright: ignore[reportUnknownMemberType]
					released_date.string = sku_element.attrs['title']
					description_element.insert_before(released_date)

				if price := product_info.select_one('.price'):
					for sibling in price.find_next_siblings():
						if isinstance(sibling, Tag) and not ((bundle := sibling.find(class_='bundle')) and bundle and bundle.get_text().startswith('Products in this bundle:')):
							sibling.decompose()
					price.decompose()

				html.append(product_info)

				images = product_listing.select_one('div.image-navigation') or soup.new_tag('div')	# pyright: ignore[reportUnknownMemberType]
				for a in images.find_all('a'):
					a.replace_with(soup.new_tag('img', attrs={'src': a.attrs['href'], 'title': a.attrs['title'], 'alt': a.attrs['title']}))  # pyright: ignore[reportUnknownMemberType]

				if product_content:
					panes = product_content.select('.product-panes > .pane > h1')
					for pane in panes:
						if pane.get_text() in ('Downloads', 'Accessories', 'Related Products'):
							if isinstance(parent := pane.parent, Tag):
								if isinstance(next_sibling := parent.find_next_sibling(), Tag) and next_sibling.name == 'hr':
									next_sibling.decompose()
								parent.decompose()

					html.append(product_content)

				elif product_content_old:
					img = soup.new_tag('img', attrs={'src': product_image})  # pyright: ignore[reportUnknownMemberType]
					if img.attrs['src'] not in (i.attrs['href'] for i in images.find_all('a')):
						html.append(img)
					html.append(images)

					for pane_title in ('downloads', 'related-products', 'accessories', 'others-also-purchased'):
						if pane := product_content_old.select_one(f'.pane.{pane_title}'):
							pane.decompose()

					html.append(product_content_old)

				product_image = urlopen(Request(product_image, headers={'User-Agent': 'Mozilla'})) if product_image else None

				return PackageData(
					global_id = global_id,
					prefix = _STORE_PREFIX,
					store = _STORE_NAME,
					sku = sku,
					name = name,
					artists = artists,
					description = description,
					image = product_image,
					readme = HTML(string=html.decode_contents(), base_url=base_url).write_pdf(stylesheets=[stylesheet])#, presentational_hints=True)
				)

		return PackageData()


import zlib
import warnings

from re import findall, search
from ssl import SSLContext
from urllib.parse import ParseResult, urljoin, urlparse
from urllib.request import urlopen, Request
from uuid import uuid5, NAMESPACE_URL

from bs4 import BeautifulSoup, Tag, XMLParsedAsHTMLWarning
from weasyprint import CSS, HTML
from weasyprint.urls import StreamingGzipFile

from . import Scraper
from ..generator import PackageData

class ShareCG(Scraper):
	domain = 'sharecg.com'

	# ShareCG requires known User-Agent, does not recognize WeasyPrint, so override
	@staticmethod
	def mozilla_url_fetcher(url: str, timeout: int = 10, ssl_context: SSLContext | None = None):
		HTTP_HEADERS = {
			'User-Agent': 'Mozilla',
			'Accept': '*/*',
			'Accept-Encoding': 'gzip, deflate',
		}

		response = urlopen(Request(url, headers=HTTP_HEADERS), timeout=timeout, context=ssl_context)
		response_info = response.info()
		result = {
			'redirected_url': response.geturl(),
			'mime_type': response_info.get_content_type(),
			'encoding': response_info.get_param('charset'),
			'filename': response_info.get_filename(),
		}
		content_encoding = response_info.get('Content-Encoding')
		if content_encoding == 'gzip':
			result['file_obj'] = StreamingGzipFile(fileobj=response)
		elif content_encoding == 'deflate':
			data = response.read()
			try:
				result['string'] = zlib.decompress(data)
			except zlib.error:
				# Try without zlib header or checksum
				result['string'] = zlib.decompress(data, -15)
		else:
			result['file_obj'] = response
		return result

	@staticmethod
	def scrape(url: ParseResult) -> PackageData:
		_STORE_NAME = 'ShareCG'
		_STORE_PREFIX = 'SHARECG'

		request = urlopen(Request(url.geturl(), headers={'User-Agent': 'Mozilla'}))
		actual_url = urlparse(str(request.url))

		# ShareCG serves broken XHTML, so xml parser does not work. Use lxml (HTML) parser instead and suppress warning
		with warnings.catch_warnings():
			warnings.simplefilter('ignore', XMLParsedAsHTMLWarning)
			soup = BeautifulSoup(request.read(), 'lxml')

		base_url = base_tag.attrs['href'] if (base_tag := soup.select_one('base[href]')) else f'{ actual_url.scheme }://{ actual_url.netloc }'
		canonical_url = url_element.attrs['content'].partition('browse/')[0] if isinstance(url_element := soup.find('meta', attrs={'name': 'twitter:url'}), Tag) else None

		if canonical_url:
			html = BeautifulSoup('', 'lxml')
			stylesheet = CSS(string='@page { margin: 1em; } img { max-width: 100%; }')

			global_id = uuid5(NAMESPACE_URL, canonical_url)
			sku = int(sku_match.group(1)) if (sku_match := search(r'(\d+)', canonical_url)) else None

			if content_html := soup.select_one('#topHalf'):
				name_element = content_html.select_one('div.ViewUploadTitle')
				product_image_url = findall(r'\d+', string=a_element.attrs['onclick'])[0] if (a_element := content_html.select_one("a.dialog_link_view_WebPage_img")) and 'onclick' in a_element.attrs else None
				if not product_image_url:
					product_image_url = img.attrs['src'] if (img := content_html.select_one("#imagefocus img")) and img.attrs['src'] else None

				if desc_element := content_html.select_one('#moreUploadInfo'):
					del desc_element.attrs['style']
					if a := desc_element.select_one('a:last-child'):
						a.decompose()
				else:
					desc_element = content_html.select_one('#moreUploadInfoExpander')

				name = name_element.get_text() if name_element else None
				description = desc_element.get_text() if desc_element else None
				artists = [artist_element.get_text().rpartition(' ')[0]] if (artist_element := content_html.select_one('a.headerLink')) else []

				if name_element:
					html.append(name_element)

				if artists:
					by = soup.new_tag('div')  # pyright: ignore[reportUnknownMemberType]
					by.string = f'By: {artists[0]}'
					html.append(by)

				html.append(soup.new_tag('hr'))	# pyright: ignore[reportUnknownMemberType]

				if desc_element:
					html.append(desc_element)

				html.append(soup.new_tag('hr'))	# pyright: ignore[reportUnknownMemberType]

				if metadata := content_html.select_one('span.uploadInfoLists'):
					html.append(metadata)

				if product_image_url:
					img = soup.new_tag('img')  # pyright: ignore[reportUnknownMemberType]
					img.attrs['src'] = f'/get_image.php?upload_image_id={ product_image_url }' if product_image_url.isdigit() else product_image_url
					html.append(img)
					product_image = urlopen(Request(urljoin(base_url, img.attrs['src']), headers={'User-Agent': 'Mozilla'}))

					if images := content_html.select('div.imageChooser a'):
						for id in [findall(r'\d+', x.attrs['onclick'])[0] for x in images]:
							if id != product_image_url:
								img = soup.new_tag('img')  # pyright: ignore[reportUnknownMemberType]
								img.attrs['src'] = f'/get_image.php?upload_image_id={ id }'
								html.append(img)
				else:
					product_image = None

				return PackageData(
					global_id = global_id,
					prefix = _STORE_PREFIX,
					store = _STORE_NAME,
					sku = sku,
					name = name,
					artists = artists,
					description = description,
					image = product_image,
					readme = HTML(string=html.decode_contents(), base_url=base_url, url_fetcher=ShareCG.mozilla_url_fetcher).write_pdf(stylesheets=[stylesheet])
				)

		return PackageData()


import re

from argparse import ArgumentParser, ArgumentTypeError
from pathlib import Path
from urllib.parse import urlparse, ParseResult
from uuid import UUID

from .generator import PackageGenerator
from .scrapers.scraper import get_scraper

def _content_location(path: str) -> Path:
	try:
		if (resolved_path := Path(path).resolve(True)).is_dir():
			return resolved_path
		else:
			raise ValueError(f'{path} is not a directory')
	except Exception as e:
		raise ArgumentTypeError(e)

def _url(url_str: str) -> ParseResult:
	url = urlparse(url_str if "://" in url_str else "http://" + url_str)
	if not all((url.scheme, url.netloc)):
		raise ArgumentTypeError('Invalid URL')

	try:
		get_scraper(url)
	except ValueError as e:
		raise ArgumentTypeError(e)

	return url

def _prefix(store: str) -> str:
	if store in (reserved := ('IM', 'DZ', 'DAZ', 'DAZ3D', 'TAFI')):
		raise ArgumentTypeError(f'Invalid prefix: {reserved} are reserved')	

	pattern = r'[A-Z][0-9A-Z]{0,6}'
	if not re.compile(pattern).fullmatch(store):
		raise ArgumentTypeError(f'Invalid prefix: Must match regex: {pattern}')

	return store

def _sku(sku: str) -> int:
	try:
		int_sku = int(sku)
		if int_sku > 99999999:
			raise ValueError
		return int_sku
	except ValueError:
		raise ArgumentTypeError(f'Invalid SKU: Must be 1 - 8 digits')

def _id(id: str) -> int:
	try:
		int_id = int(id)
		if int_id > 99:
			raise ValueError
		return int_id
	except ValueError:
		raise ArgumentTypeError(f'Invalid ID: Must be 1 - 2 digits')

def _readme(file: str) -> Path:
	readme = Path(file)
	if not (readme.is_file() and readme.suffix.lower() == '.pdf'):
		raise ArgumentTypeError('ReadMe file must be a PDF')

	return readme

def _main() -> None:
	parser = ArgumentParser(description = 'Generate DAZ Studio Content Package')
	parser.add_argument('content_location', type = _content_location, help = 'Content Directory')
	parser.add_argument('-g', '--global-id', type = UUID, help = 'Product Global ID')
	parser.add_argument('-u', '--url', type = _url, help = 'URL for product information')
	parser.add_argument('-p', '--prefix', type = _prefix, help = 'Source Prefix')
	parser.add_argument('-s', '--sku', type = _sku, help = 'Product SKU')
	parser.add_argument('-I', '--id', type = _id, default = '0', help = 'Package ID')
	parser.add_argument('-S', '--store', type = str, help = 'Product Store')
	parser.add_argument('-n', '--name', type = str, help = 'Product [Part] Name')
	parser.add_argument('-t', '--tags', type = str, nargs = '+', help = 'Tags')
	parser.add_argument('-a', '--authors', type = str, nargs = '+', help = 'Authors')
	parser.add_argument('-d', '--description', type = str, help = 'Product Description')
	parser.add_argument('-i', '--image', type = Path, help = 'Product Image')
	parser.add_argument('-r', '--readme', type = _readme, help = 'Product ReadMe')
	parser.add_argument('-v', '--verbose', action = 'store_true', help = 'enable verbose output')

	args = parser.parse_args()

	package_generator = PackageGenerator(
		global_id=args.global_id,
		prefix=args.prefix,
		sku=args.sku,
		store=args.store,
		name=args.name,
		tags=args.tags,
		artists=args.authors,
		description=args.description,
		image=args.image,
		readme=args.readme,
		id=args.id,
		url=args.url,
		content_location=args.content_location,
		verbose=args.verbose)
	
	package_generator.make_package()

def _profile() -> None:  # pyright: ignore[reportUnusedFunction]
	import cProfile, pstats
	profiler = cProfile.Profile()
	profiler.enable()
	_main()
	profiler.disable()
	pstats.Stats(profiler).dump_stats('stats.profiler')
 
if __name__ == '__main__':
	# _profile()
	_main()